
from typing import Sequence

import numpy as np
from numpy.typing import NDArray
from scipy.interpolate import LinearNDInterpolator

Pair = tuple[float, float]
FArray = NDArray[np.float64]


def random_weights(n: int = 3) -> FArray:
    """Get n>=2 random weights that sum ≈ 1. Shape will be (n,1)."""
    if n < 2:
        raise ValueError("n must be at least 2")
    # python
    # weights = [random.random() for _ in range(n)]
    # total = sum(weights)
    # weights = [[w/total] for w in weights]

    # numpy
    rng = np.random  # compat with older
    # rng = np.random.default_rng() # new style
    weights = rng.uniform(size=n)
    weights = weights/weights.sum()
    return weights.reshape((n, 1))


def create_verts(n: int = 3, bounds: Sequence[Pair] | None = None) -> FArray:
    """Create `n` vertices according to the bounds. Can have more
    than 3 dimensions at each point. Will produce a (n , len(bounds)) 
    shaped array. Defaults to creating 3 vertices of ℝ³ ∈ [0,1)"""
    if n < 2:
        raise ValueError("requires at least 2 verts")
    if not bounds:
        bounds = [(0, 1)]*3
    if len(bounds) < 3:
        raise ValueError("requires at least 3 dimensions")

    rng = np.random
    # rng = np.random.default_rng() # new style
    dimensions = [rng.uniform(b[0], b[1], size=n) for b in bounds]
    return np.vstack(dimensions).transpose()


def point_in_triangle(triangle: FArray, weights: FArray) -> FArray:
    """Calculate the point in triangle according to the weights.
    Points will have shape (len(weights), ). Expected that 
    `triangle.shape[0] == weights.shape[0]`"""
    if triangle.shape[0] != weights.shape[0]:
        raise ValueError("triangle vertex dimensions do not match number of weights")

    point = (triangle*weights).sum(axis=0)
    return point


def interpolate_z(triangle: FArray, point: FArray) -> FArray:
    """Run LinearNDInterpolator using the xy coords (slice 0:2) of 3+ dimension
    triangle and the xy coords (slice 0:2) of point. Returns just the interpolated 
    values (slice 2:) at point. Returned array will have 
    shape (1, triangle.shape[1] - 2), aka number of dimensions of each vertex
    minus 2 (for x and y)"""
    interp = LinearNDInterpolator(triangle[:, :2], triangle[:, 2:])
    z = interp(point[:2])
    return z


def calc_barycentric_weights(triangle: FArray, point: FArray) -> FArray:
    """Calculates the barycentric coordinates of point (shape (N-dims,)) with
    respect to triangle (shape (3, N-dims))"""
    # https://en.wikipedia.org/wiki/Barycentric_coordinate_system#Edge_approach
    D = 2
    r1mr3 = triangle[0, :D]-triangle[2, :D]
    r2mr3 = triangle[1, :D]-triangle[2, :D]
    r3mr1 = triangle[2, :D]-triangle[0, :D]
    pmr3 = point[:D]-triangle[2, :D]

    # stupid np deprecating 2d cross product
    # https://github.com/numpy/numpy/issues/26620
    cross = np.cross
    if D == 2:
        def cross(a, b): return a[0]*b[1] - a[1]*b[0]

    denom = cross(r1mr3, r2mr3)
    w1 = cross(pmr3, r2mr3)/denom
    w2 = cross(pmr3, r3mr1)/denom
    w3 = 1-w1-w2
    # w1-3 will each be shape(3,) of the same value for D=3
    # but will be scalars for D=2. flat[0] works for both cases
    return np.array((w1.flat[0], w2.flat[0], w3.flat[0])).reshape((3, 1))


def run_trial():
    x_bounds = (100_000, 100_100)
    y_bounds = (3_000_000, 3_000_100)
    z_bounds = (100, 200)
    t = create_verts(bounds=(x_bounds, y_bounds, z_bounds, (-10, -20)))
    # w = np.repeat(1/3, 3) # center
    w = random_weights()
    p = point_in_triangle(t, w)
    z = interpolate_z(t, p)
    w_calc = calc_barycentric_weights(t, np.hstack((p[:2], z[0])))
    with np.printoptions(formatter={'float_kind': "{:.6f}".format}):
        print("triangle\n", t)
        print("rand weights\n", w)
        print("point\n", p)
        # print(t[:, :2])
        # print(t[:, 2])
        print("p_xy, interp z\n", p[:2], z)
        print("weights with interp z\n", w_calc)

    assert np.isclose(p[2:], z[0]).all()
    assert np.isclose(w, calc_barycentric_weights(t, p)).all()
